open Tools

(** { 1 } Types. *)

(** Type d'un automate déterministe avec des états de type ['a] sur des lettres de 
    type ['b]. *)
type ('a,'b) auto_min = {
  etats : 'a list;
  initial : 'a ;
  alphabet : 'b list;
  final : 'a -> bool;
  trans : 'a -> 'b -> 'a 
}

(** Type d'un mot sur un alphabet de type ['a]. *)
type 'a mot = 'a list

(** Exécution d'un automate déterministe sur un mot. *)
let exec a (u : 'a mot) = 
  List.fold_left a.trans a.initial u

(** Type d'un automate avec des états de type ['a] sur des lettres de type ['b]. 
    Un tel automate est vu comme un tuple [(e,i,f,a,t)] où :
    - [e] est l'ensemble des états, 
    - [i] est l'ensemble des états initiaux, 
    - [f] est l'ensemble des états finaux, 
    - [a] est l'alphabet, 
    - [tr] est l'ensemble des transitions *)
type ('a,'b) automate = 
    ('a list * 'a list * 'a list * 'b list * ('a * 'b * 'a) list) 


(** { 1 } Opérations sur les automates. *)

(** Application d'une fonction de renommage à tous les états d'un automate. *)
let map (f : 'a -> 'b) (e,i,fn,a,tr : ('a,'c) automate) : ('b,'c) automate =
  (List.map f e,List.map f i,List.map f fn, a,
   List.map (fun (q,a,q') -> (f q,a,f q')) tr) 

(** Renommage des états d'un automate avec des entiers. *)
let simpl (e,_,_,_,_ as a : ('a,'b) automate) : (int,'b) automate =
  let (conv1,_) = makebij e in
  map conv1 a

(** Construction d'un automate à partir d'une fonction de transition. *)
let build 
    (type etat)
    (module E : Base with type t = etat)
    (init : etat)
    (final : etat -> bool)
    (alphabet : 'b list)
    (trans : etat -> 'b -> etat)
    : (etat,'b) automate =
  let c = E.compare in
  let rec step etats tr = function
    | [] -> (etats,tr)
    | q::todo -> 
      let nouv,tr' = 
	mappairs 
	  c
	  (fun a -> let q' = trans q a in (q',(q,a,q')))
	  tr
	  alphabet 
      and etats' = inser c q etats
      in
      step 
	etats'
	tr' 
	(conc c [List.filter (fun q' -> not (mem c q' etats')) nouv;todo])
  in
  let etats,tr = step [] [] [init] in
  let fn = List.filter final etats in
  (etats,[init],fn,alphabet,tr)

(** Construction d'un automate à partir d'une fonction de transition, 
    en omettant l'état puit. *)
let build_avec_puit 
    (type etat)
    (module E : Base with type t = etat)
    (init : etat)
    (final : etat -> bool)
    (alphabet : 'b list)
    (trans : etat -> 'b -> etat)
    (puit : etat)
    : (etat,'b) automate =
  let c = E.compare in
  let rec step etats tr = function
    | [] -> (etats,tr)
    | q::todo -> 
      let etats' = inser c q etats in
      let todo',tr' =
	List.fold_left 
	  (fun (acc1,acc2) a ->
	    let q' = trans q a in
	    if q' = puit
	    then (acc1,acc2)
	    else 
	      if (mem c q' etats')
	      then (acc1,(q,a,q')::acc2)
	      else (inser c q' acc1,(q,a,q')::acc2))
	  (todo,tr)
	  alphabet 
      in
      step etats' tr' todo'
  in
  let etats,tr = step [] [] [init] in
  let fn = List.filter final etats in
  (etats,[init],fn,alphabet,tr)

(** Construction d'un automate à partir d'une fonction de transition. *)
let build2 
    (type etat)
    (module E : Base with type t = etat)
    (init : etat)
    (final : etat -> bool)
    (alphabet : 'b list)
    (trans : etat -> 'b -> etat list)
    : (etat,'b) automate =
  let c = E.compare in
  let rec step etats tr = function
    | [] -> (etats,tr)
    | q::todo -> 
      let etats' = inser c q etats in
      let todo',tr' =
	List.fold_left 
	  (fun (acc1,acc2) a ->
	    List.fold_left 
	      (fun (acc1,acc2) q' ->
		if (mem c q' etats')
		then (acc1,(q,a,q')::acc2)
		else (inser c q' acc1,(q,a,q')::acc2))
	      (acc1,acc2)
	      (trans q a))
	  (todo,tr)
	  alphabet 
      in
      step etats' tr' todo'
  in
  let etats,tr = step [] [] [init] in
  let fn = List.filter final etats in
  (etats,[init],fn,alphabet,tr)

(** Construction d'un automate à partir d'une fonction de transition 
    utilisant un accumulateur. *)
let build_acc
    (type etat)
    (module E : Base with type t = etat)
    (init : etat)
    (final : etat -> bool)
    (alphabet : 'a list)
    (trans : 'b -> etat -> 'a -> 'b * etat)
    (acc0 : 'b)
    : (etat,'a) automate =
  let c = E.compare in
  let rec step etats tr acc = function
    | [] -> (etats,tr,acc)
    | q::todo -> 
      let etats' = inser c q etats in
      let todo',tr',acc' =
	List.fold_left 
	  (fun (acc1,acc2,acc3) a ->
	    let acc3',q' = trans acc3 q a in
	    if (mem c q' etats')
	    then (acc1,(q,a,q')::acc2,acc3')
	    else (inser c q' acc1,(q,a,q')::acc2,acc3'))
	  (todo,tr,acc)
	  alphabet 
      in
      step etats' tr' acc' todo'
  in
  let etats,tr,_ = step [] [] acc0 [init] in
  let fn = List.filter final etats in
  (etats,[init],fn,alphabet,tr)

(** Construction d'un automate à partir d'une fonction de transition. *)
let build2_acc 
    (type etat)
    (module E : Base with type t = etat)
    (init : etat list)
    (final : etat -> bool)
    (alphabet : 'b list)
    (trans : 'c -> etat -> 'b -> 'c * etat list)
    (acc0 : 'c)
    : (etat,'b) automate =
  let c = E.compare in
  let rec step etats tr acc = function
    | [] -> (etats,tr,acc)
    | q::todo -> 
      let etats' = inser c q etats in
      let todo',tr',acc' =
	List.fold_left 
	  (fun (acc1,acc2,acc3) a ->
	    let acc3',ql = trans acc3 q a in
	    let acc1',acc2' = 
	      List.fold_left 
		(fun (acc1,acc2) q' ->
		  if (mem c q' etats')
		  then (acc1,(q,a,q')::acc2)
		  else (inser c q' acc1,(q,a,q')::acc2))
		(acc1,acc2)
		ql
	    in
	    acc1',acc2',acc3')
	  (todo,tr,acc)
	  alphabet 
      in
      step etats' tr' acc' todo'
  in
  let etats,tr,_ = step [] [] acc0 (List.sort c init) in
  let fn = List.filter final etats in
  (etats,init,fn,alphabet,tr)

(** Automate mirroir. *)
let rev (e, i, f, alph, tr : ('a,'b) automate) : ('a,'b) automate =
  (e, f, i, alph, List.map (fun (q,a,q') -> (q',a,q)) tr)

(** Déterminisation. *)
let determ 
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat) 
    (module Alph : Base with type t = lettre) 
    (e,i,fn,a,tr : (etat,lettre) automate) =
  let tr' = 
    let tr'= makefuncnd (module Etat)(module Alph) tr 
    in
    fun ql a -> 
      List.fold_left 
	(fun acc q -> 
	  List.fold_left 
	    (fun acc q' -> inser (Etat.compare) q' acc) 
	    acc 
	    (tr' q a))
	[] 
	ql
(*
bind Etat.compare (fun q -> tr' q a) ql*)
  in
  let fn' = List.sort Etat.compare fn in
  (build (module ListB(Etat)) i (fun q -> inter Etat.compare q fn') a tr')

(** Déterminisation, sans état puit. *)
let determ2 
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat) 
    (module Alph : Base with type t = lettre) 
    (e,i,fn,a,tr : (etat,lettre) automate) =
  let tr' = 
    let tr'= makefuncnd (module Etat)(module Alph) tr 
    in
    fun ql a -> bind Etat.compare (fun q -> tr' q a) ql
  in
  let fn' = List.sort Etat.compare fn in
  (build_avec_puit (module ListB(Etat)) i (fun q -> inter Etat.compare q fn') a tr' [])

(** Minimisation. *)
let min
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat) 
    (module Alph : Base with type t = lettre) 
    aut = 
  determ (module ListB(Etat))(module Alph) 
    (rev (determ (module Etat)(module Alph) (rev aut)))

(** Minimisation, sans état puit. *)
let min2
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat) 
    (module Alph : Base with type t = lettre) 
    aut = 
  determ2 (module ListB(Etat))(module Alph) 
    (rev (determ2 (module Etat)(module Alph) (rev aut)))

(** Conversion vers un automate minimal fonctionnel *)
let convert
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat) 
    (module Alph : Base with type t = lettre) 
    ((e,i,f,a,tr) as aut : (etat,lettre) automate) =
  let (q0,e',f',t') = 
      match min (module Etat) (module Alph) aut with
	| (e,[q0],f,_,t) -> (q0,e,f,t)
	| _ -> failwith "error error xwzcwxfzzwxyyxwwz..." 
    in
    let e'' = None::(List.map (fun q -> Some q) e') in
    let module LLE = ListB(ListB(Etat)) in
    let final = function 
      | None -> false 
      | Some q -> mem LLE.compare q f' in
    let trans =
      let f = makefuncnd (module LLE)(module Alph) t' in
      (function 
	| None -> 
	  (fun a -> None)
	| Some q -> 
	  (fun a -> 
	    match (f q a) with 
	      | [] -> None
	      | [q'] -> Some q' 
	      | _ -> failwith "wxfzzwxyyxwwz!!!"))
    in
    {etats=e'';initial=Some q0;alphabet=a;final = final;trans=trans}

(** Test simple d'équivalence d'automates *)
let equiv 
    (type etat1)
    (type etat2)
    (type lettre)
    (module Etat1 : Base with type t = etat1) 
    (module Etat2 : Base with type t = etat2) 
    (module Alph : Base with type t = lettre) 
    ((e1,i1,fn1,a1,tr1) : (etat1,lettre) automate)
    ((e2,i2,fn2,a2,tr2) : (etat2,lettre) automate) 
    : lettre mot option =  
  let alph = a1 ^^ a2 in
  let module E = struct exception Differents of lettre mot end in
  let tr1,tr2 =
    let tr1 = makefuncnd (module Etat1)(module Alph) tr1
    and tr2 = makefuncnd (module Etat2)(module Alph) tr2 in
    ((fun ql a -> 
      bind Etat1.compare (fun q -> try tr1 q a with Not_found -> []) ql),
     fun ql a -> 
       bind Etat2.compare (fun q -> try tr2 q a with Not_found -> []) ql)
  in
  let i1,i2 = (List.sort Etat1.compare i1,List.sort Etat2.compare i2) in
  let fn1,fn2 = (List.sort Etat1.compare fn1,List.sort Etat2.compare fn2) in
  let is_fn1,is_fn2 = (inter Etat1.compare fn1,inter Etat2.compare fn2) in
  let comp = 
    comp_to_pair (comp_to_list Etat1.compare) (comp_to_list Etat2.compare) 
  in
  let comp' = 
    comp_to_tripl (comp_to_list Alph.compare) (comp_to_list Etat1.compare) (comp_to_list Etat2.compare) 
  in
  let step pairs (u,ql1,ql2) =
    if ((is_fn1 ql1) = (is_fn2 ql2))
    then
      begin
	let next = List.map (fun a -> (a::u,tr1 ql1 a,tr2 ql2 a)) alph in
	List.filter (fun (_,q1,q2) -> not (mem comp (q1,q2) pairs)) next
      end 
    else 
      (raise (E.Differents u))
  in
  let rec aux pairs = function
    | [] -> None
    | ((_,q1,q2) as p)::todo ->
      try
	let pairs' = (inser comp (q1,q2) pairs) in
	aux pairs' (conc comp' [step pairs' p;todo])
      with
	| E.Differents u -> (Some (List.rev u))
  in
  aux [] [[],i1,i2]

(** Algorithme d'Hopcroft and Karp *)
let hkequiv
    (type etat1)
    (type etat2)
    (type lettre)
    (module Etat1 : Base with type t = etat1) 
    (module Etat2 : Base with type t = etat2) 
    (module Alph : Base with type t = lettre) 
    ((e1,i1,fn1,a1,tr1) : (etat1,lettre) automate)
    ((e2,i2,fn2,a2,tr2) : (etat2,lettre) automate) 
    : lettre mot option =
  let alph = a1 ^^ a2 in
  let module E = struct exception Differents of lettre mot end in
  let tr1,tr2 =
    let tr1 = makefuncnd (module Etat1)(module Alph) tr1
    and tr2 = makefuncnd (module Etat2)(module Alph) tr2 in
    ((fun ql a -> 
      bind Etat1.compare (fun q -> try tr1 q a with Not_found -> []) ql),
     fun ql a -> 
       bind Etat2.compare (fun q -> try tr2 q a with Not_found -> []) ql)
  in
  let i1,i2 = (List.sort Etat1.compare i1,List.sort Etat2.compare i2) in
  let fn1,fn2 = (List.sort Etat1.compare fn1,List.sort Etat2.compare fn2) in
  let is_fn1,is_fn2 = (inter Etat1.compare fn1,inter Etat2.compare fn2) in 
  let module UF = MakeUF(ListB(Etat1))(ListB(Etat2)) in
  let step state u ql1 ql2 todo =
    if UF.equivalent (A ql1) (B ql2) state
    then todo
    else
      if ((is_fn1 ql1) = (is_fn2 ql2))
      then (List.fold_left (fun acc a -> (a::u,tr1 ql1 a,tr2 ql2 a)::acc) todo alph)
      else (raise (E.Differents u))
  in
  let rec aux state = function
    | [] -> None
    | (u,ql1,ql2)::todo ->
      try
	let next = step state u ql1 ql2 todo in
	let (state',_) = UF.union (A ql1) (B ql2) state () in
	aux state' next
      with
	| E.Differents u -> (Some (List.rev u))
  in
  aux UF.initial [[],i1,i2]

let access 
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat) 
    (module Alph : Base with type t = lettre) 
    (e,i,fn,alph,tr : (etat,lettre) automate) 
    : (etat,lettre) automate  =
  let trans = makefuncnd (module Etat) (module Alph) tr in
  let (|>) = inser Etat.compare in
  let is_in l x = mem Etat.compare x l in
  let step dejavus todo =
    let aux acc q =
      List.fold_left
	(fun acc a -> 
	  List.fold_left
	    (fun (acc1,acc2) q ->
	      match inserb Etat.compare q acc1 with
		| None -> (acc1,acc2)
		| Some acc1' -> (acc1',q|>acc2))
	    acc
	    (trans q a))
	acc
	alph
    in
    List.fold_left aux (dejavus,[]) todo
  in
  let rec aux dejavus todo =
    let dejavus',todo' = step dejavus todo in
    if todo' = []
    then dejavus'
    else aux dejavus' todo'
  in
  let i' = List.sort Etat.compare i in
  let e' = aux i' i' in
  let tr' = List.filter (fun (q,_,q') -> (is_in e' q) & (is_in e' q')) tr in
  let fn' = List.filter (fun q -> is_in e' q) fn in
  (e',i',fn',alph,tr')

(** Nettoyage de l'automate, en ne gardant que les états co-accessibles. *)
let coaccess 
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat) 
    (module Alph : Base with type t = lettre) 
    (aut : (etat,lettre) automate) 
    : (etat,lettre) automate  =
  rev (access (module Etat)(module Alph) (rev aut))

let puits 
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat)  
    (e,i,fn,a,tr : (etat,lettre) automate) 
    : etat list =
  let (--) = remove Etat.compare in
  let init = 
    List.fold_left
      (fun acc q -> q -- acc)
      (List.sort Etat.compare e)
      fn
  in
  let rec aux acc = function
    | [] -> acc
    | (q,_,q')::tr' when q=q' -> aux acc tr'
    | (q,_,_)::tr' -> aux (q--acc) tr'
  in
  aux init tr

let etats_coaccessibles
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat) 
    (module Alph : Base with type t = lettre) 
    a =
  let (e,_,_,_,_) = coaccess (module Etat) (module Alph) a in
  e
  
let etats_non_puit
    (type etat)
    (type lettre)
    (module Etat : Base with type t = etat) 
    (e,i,fn,a,tr) =
  let (|>) = inser Etat.compare in
  let fn' = List.sort Etat.compare fn in
  let rec aux acc = function
    | [] -> acc
    | (q,_,q')::tr' when q=q' -> aux acc tr'
    | (q,_,_)::tr' -> aux (q|>acc) tr'
  in
  aux fn' tr

(** { 1 } Traduction en format dot. *)

let sprint_trans s
    (conve : 'a -> string) 
    (convl : 'b -> string) 
    (q,a,q' : 'a * 'b * 'a) =
  (s^(Printf.sprintf "%s -> %s [label = \"%s\"];\n" (conve q) (conve q') (convl a)))

let sprint_states s
    (e_to_str : 'a -> string) 
    (if_i : 'a -> string) 
    (if_f : 'a -> string) 
    (e : 'a list) =
  let conve =
    let (c,_) = makebij e
    in
    (fun q -> string_of_int (c q))
  in
  (List.fold_left
     (fun s q -> 
       s^(Printf.sprintf
	    "%s [label = \"%s\"%s%s];\n" 
	    (conve q)
	    (e_to_str q)
	    (if_i q)
	    (if_f q)))
     s
     e,
   conve)

(** Impression comme chaîne de caractères. *)
let sprint_auto
    (e_to_str : 'a -> string) 
    (l_to_str : 'b -> string)
    (e,i,fn,a,tr : ('a,'b) automate) =
  let if_i q =
    if List.mem q i
    then ", shape = box"
    else ""
  and if_f q = 
    if List.mem q fn
    then ", peripheries = 2"
    else ""
  in
  let s,conve = 
    sprint_states (Printf.sprintf "digraph A {\n") e_to_str if_i if_f e 
  in
  (List.fold_left (fun s t -> sprint_trans s conve l_to_str t) s tr)^
  (Printf.sprintf "}\n")
  
let print_trans ch 
    (conve : 'a -> string) 
    (convl : 'b -> string) 
    (q,a,q' : 'a * 'b * 'a) =
  Printf.fprintf ch "%s -> %s [label = \"%s\"];\n" (conve q) (conve q') (convl a)

let print_states ch
    (e_to_str : 'a -> string) 
    (if_i : 'a -> string) 
    (if_f : 'a -> string) 
    (e : 'a list) =
  let conve =
    let (c,_) = makebij e
    in
    (fun q -> string_of_int (c q))
  in
  List.iter
    (fun q -> 
      Printf.fprintf ch 
	"%s [label = \"%s\"%s%s];\n" 
	(conve q)
	(e_to_str q)
	(if_i q)
	(if_f q))
    e;
  conve
  
(** Impression dans un canal. *)
let print_auto ch
    (e_to_str : 'a -> string) 
    (l_to_str : 'b -> string)
    (e,i,fn,a,tr : ('a,'b) automate) =
  let if_i q =
    if List.mem q i
    then ", shape = box"
    else ""
  and if_f q = 
    if List.mem q fn
    then ", peripheries = 2"
    else ""
  in
  Printf.fprintf ch "digraph A {\n";
  let conve = print_states ch e_to_str if_i if_f e in
  List.iter (print_trans ch conve l_to_str) tr;
  Printf.fprintf ch "}\n"
  
(** Impression dans un fichier. *)
let print_to_file name printetat printlettre auto = 
  let ch = open_out (name^".dot")
  in 
  print_auto ch printetat printlettre auto;
  close_out ch
