(** Monoides *)

(** Type des monoides dont les éléments sont des mots sur un alphabet de type ['a], 
    consitués d'une liste d'éléments et d'une liste de règles de réécriture. *)
type ('a,'b) monoid =
    ('a Automate.mot * 'b array) list * ('a Automate.mot * 'a Automate.mot) list

(** Impression d'un mot, étant donnée un fonction d'impression des 
    lettres. *)
val mot_to_string : ('a -> string) -> 'a Automate.mot -> string

val list_to_string : ('a -> string) -> 'a list -> string

(** Impression d'un tableau de liste, étant donnée un fonction 
    d'impression des éléments de base. *)
val table_to_string : ('b -> string) -> 'b array -> string

(** Impression d'une liste de mots, étant donnée un fonction d'impression des 
    lettres. *)
val lang_to_string : ('a -> string) -> 'a Automate.mot list -> string

(** Impression d'une position. *)
val pos_to_string : ('a -> string) -> ('a * int) option -> string

(** Résidu d'une liste de règles de réécriture par une lettre, lue par à
    gauche des membres de gauche. *)
val ( -| ) :
  'a -> ('a list * 'b) list -> (('a list * 'b) list, 'b) Tools.case

(** Une étape de réduction d'un mot par rapport à un ensemble de règles de 
    réécriture. *)
val beta : ('a list * 'a list) list -> 'a Automate.mot -> 'a Automate.mot

(** Calculs divers. *)
module Compute :
  functor (Etat : Tools.Base) ->
    functor (Alph : Tools.Base) ->
      sig
        type etat = Etat.t
        type lettre = Alph.t
        type mot = lettre list

	(** Calcul du monoide syntaxique à partir d'un automate, avec impression
	    de la  table des tranformation des états induites par chacun des 
	    éléments du monoide, et des règles de réécriture inférées. *)
        val monsyn :
	  (etat list) ->
          ((lettre -> string)*(etat -> string)) option ->
          (etat, lettre) Automate.automate -> 
	  mot list * mot list * (lettre,etat option) monoid

	(** Calcul du monoide syntaxique à partir d'un automate, avec impression
	    de la  table des tranformation des états induites par chacun des 
	    éléments du monoide, et des règles de réécriture inférées. *)
        val monsyn2 :
          ((lettre -> string)*(etat -> string)) option ->
          (etat, lettre) Automate.automate ->
	  mot list * mot list * (lettre,etat option) monoid

	(** Calcul du monoide syntaxique à partir d'un automate, avec impression
	    de la  table des tranformation des états induites par chacun des 
	    éléments du monoide, et des règles de réécriture inférées. *)
        val monsyn3 :
          (etat, lettre) Automate.automate ->
	  (int -> etat) * (lettre,etat option) monoid

	(** Réduction d'un mot en un élément du monoide. *)
        val clbeta : (lettre,'a) monoid -> mot -> mot

	(** Produit dans le monoide. *)
        val prod : (lettre,'a) monoid -> mot -> mot -> mot

	(** Produit d'un mot par un ensemble de mot. *)
        val prodl : (lettre,'a) monoid -> mot -> mot list -> mot list

	(** Produit d'un ensemble de mots par un mot. *)
        val prodr : (lettre,'a) monoid -> mot list -> mot -> mot list

	(** Produit de deux ensembles de mots. *)
        val prodlst : (lettre,'a) monoid -> mot list -> mot list -> mot list

	(** Étoile d'un ensemble de mots : [(1+A)^n], tel que 
	    [(1+A)^{n+1}=(1+A)^n]. *)
        val star : (lettre,'a) monoid -> mot list -> mot list

	(** Partie droite de la construction de Ésik et Bernátsky. *)
        val func_g :
          (lettre,'a) monoid ->
          (lettre -> lettre) -> mot list -> lettre -> mot list

	(** Partie droite de la construction de Ésik et Bernátsky. *)
        val auto_g :
          (lettre,'a) monoid ->
          mot list ->
          lettre list ->
          (lettre -> lettre) -> (mot list, lettre) Automate.automate

	(** Construction de Ésik et Bernátsky. *)
        val auto_f :
          (lettre,'a) monoid ->
          mot list ->
	  mot list ->
          lettre list ->
          (lettre -> lettre) ->
          (mot list * mot list, lettre) Automate.automate

	(** Construction de Ésik et Bernátsky optimisée. *)
	val auto_f_opt :
	  (lettre,'a) monoid ->
	  mot list ->
	  mot list ->
	  lettre list ->
	  (lettre -> lettre) ->
	  (mot list * mot list, lettre) Automate.automate

	(** Construction de Ésik et Bernátsky optimisée. *)
	val auto_f_opt2 :
	  (lettre,'a) monoid ->
	  mot list ->
	  mot list ->
	  lettre list ->
	  (lettre -> lettre) ->
	  (mot list * mot list, lettre) Automate.automate

	(** Construction de Ésik et Bernátsky optimisée. *)
	val auto_f_opt3 :
	  (lettre,'a) monoid ->
	  mot list ->
	  mot list ->
	  lettre list ->
	  (lettre -> lettre) ->
	  (mot list * mot list, lettre) Automate.automate

	(** Automate de Cayley du monoide*)
	val cayleyr : 
	  (lettre,'a) monoid -> mot list -> lettre list -> 
	  (mot, lettre) Automate.automate
      end
