module type R = sig

  type elt

  type t

  val empty : t

  val id : elt list -> t

  val equal : t -> t -> bool
    
  val compare : t -> t -> int

  val succ : elt -> t -> elt list
  
  val restr : elt list -> t -> t

  val inser : (elt*elt) -> t -> t

  val mem : (elt*elt) -> t -> bool

  val map : (elt*elt -> elt*elt) -> t -> t

  val fold : (elt*elt -> 'a -> 'a) -> t -> 'a -> 'a

  val to_list : t -> (elt*elt) list

  val concatenation : t -> t -> t

  val union : t -> t -> t

  val refl : elt list -> t -> t

  val etoile : elt list -> t -> t
end

module Make (E : Tools.Base) : R with type elt = E.t = struct

  module Map = Map.Make(E)
  module Set = Set.Make(E)

  type elt = E.t

  type t = Set.t Map.t

  let (|>) = Set.add

  let domain r =
    Map.fold (fun x s acc -> Set.union s (Set.add x acc)) r Set.empty

  let empty = Map.empty

  let id dom = 
    List.fold_left (fun acc x -> Map.add x (Set.singleton x) acc) empty dom

  let equal r1 r2 = Map.equal Set.equal r1 r2

  let compare r1 r2 = Map.compare Set.compare r1 r2

  let succ elt r =
    try Set.elements (Map.find elt r)
    with Not_found -> []

  let restr l r =
    List.fold_left 
      (fun acc x -> 
	try Map.add x (Map.find x r) acc 
	with Not_found -> acc) 
      Map.empty
      l

  let inser (x,y) (r:t) : t  = 
    try Map.add x (y|>(Map.find x r)) r
    with Not_found -> Map.add x (Set.singleton y) r
	
  let mem (x,y) r =
    try Set.mem y (Map.find x r)
    with Not_found -> false

  let map f r =
    Map.fold (fun x s r -> Set.fold (fun y r -> inser (f (x,y)) r) s r) r Map.empty

  let fold f r a =
    Map.fold (fun x s b -> Set.fold (fun y c -> f (x,y) c) s b) r a

  let to_list r =
    fold (fun x l -> x::l) r []

  let concatenation (r1:t) (r2:t) :t =
    let aux (x,y) (acc : t) : t =
      try 
	let s = Map.find y r2 in
	Set.fold (fun z r -> inser (x,z) r) s acc
      with 
	| Not_found -> acc
    in
    fold aux r1 Map.empty
    
  let union r1 r2 =
    fold inser r1 r2

  let (==) = equal

  let (@@) = concatenation

  let (++) = union

  let refl dom r =
    List.fold_left (fun acc x -> inser (x,x) acc) r dom

  let etoile dom r =
    let rec aux r =
      let r' = r@@r in
      if r==r'
      then r
      else aux r'
    in
    aux (refl dom r)

end
